//=============================================================================================================================
//
// Copyright (c) 2015-2019 VisionStar Information Technology (Shanghai) Co., Ltd. All Rights Reserved.
// EasyAR is the registered trademark or trademark of VisionStar Information Technology (Shanghai) Co., Ltd in China
// and other countries for the augmented reality technology developed by VisionStar Information Technology (Shanghai) Co., Ltd.
//
//=============================================================================================================================
#include "MyCamera.h"

#import <easyar/vector.oc.h>
#import <easyar/camera.oc.h>
#import <easyar/buffer.oc.h>
#import <easyar/image.oc.h>
#import <easyar/cameraparameters.oc.h>
#import <easyar/frame.oc.h>
#include <chrono>
#include <memory>

#import <Endian.h>
#import <UIKit/UIKit.h>

@interface MyCamera() < AVCaptureVideoDataOutputSampleBufferDelegate >
{
    easyar_CameraParameters *CameraParams;
    bool opened_;
}
@property(nonatomic, weak) AVCaptureDevice *videoDevice;
@property(nonatomic, strong) NSMutableArray *allSupportedPreviewSize;
@property(nonatomic) CGSize previewSize;
@property(nonatomic, strong) NSArray *allSupportedFPS;
@property(nonatomic, strong) AVCaptureVideoDataOutput *captureOutPut;
@end

@implementation MyCamera

- (instancetype)init
{
    self = [super init];
    
    if (self) {
        self.captureSession = [[AVCaptureSession alloc] init];
        self.captureOutPut = nil;
        opened_ = false;
    }
    return self;
}

-(void) connect:(easyar_InputFrameSink *)inputFrameSink
{
    self.sink = inputFrameSink;
}

-(BOOL)open
{
    self.videoDevice = [self cameraWithPosition : AVCaptureDevicePositionBack];
    if (!self.videoDevice) {
        return FALSE;
    }
    
    NSError *error;
    AVCaptureDeviceInput *videoInput = [AVCaptureDeviceInput deviceInputWithDevice : self.videoDevice error : &error];
    if (error) {
        return FALSE;
    }
    [self setDeviceInput : videoInput];
    
    if ([self.captureSession canAddInput : videoInput]) {
        [self.captureSession addInput : videoInput];
    } else {
        NSLog(@"Could not add input port to capture session %@", self.captureSession);
    }
    
    [self addVideoDataOutput];
    opened_ = true;
    
    self.allSupportedPreviewSize = [[NSMutableArray alloc] init];
    
    if ([self.captureSession canSetSessionPreset : AVCaptureSessionPreset1280x720]) {
        [self.allSupportedPreviewSize addObject : AVCaptureSessionPreset1280x720];
    }
    if ([self.captureSession canSetSessionPreset : AVCaptureSessionPreset640x480]) {
        [self.allSupportedPreviewSize addObject : AVCaptureSessionPreset640x480];
    }
    if ([self.captureSession canSetSessionPreset : AVCaptureSessionPreset1920x1080]) {
        [self.allSupportedPreviewSize addObject : AVCaptureSessionPreset1920x1080];
    }
    if (self.allSupportedPreviewSize.count > 0) {
        CGSize defaultSize = [self getSupportedSize : 0];
        [self setSize : defaultSize];
    }
    
    return TRUE;
}

-(BOOL)start
{
    if (!self.captureSession) {
        return FALSE;
    }
    [self.captureSession startRunning];
    return TRUE;
}

-(BOOL)stop
{
    if (!self.captureSession) {
        return FALSE;
    }
    [self.captureSession stopRunning];
    return TRUE;
}

- (easyar_CameraParameters *) CameraParams
{
    return CameraParams;
}

-(CGSize)size
{
    return self.previewSize;
}

-(int)getNumSupportedSize
{
    if (!opened_) {
        return 0;
    }
    return (int)(self.allSupportedPreviewSize.count);
}

-(CGSize)getSupportedSize:(int)idx
{
    if (!opened_ || idx < 0 || idx >[self getNumSupportedSize]) {
        return CGSizeMake(0, 0);
    }
    NSString *preset = [self.allSupportedPreviewSize objectAtIndex : idx];
    if ([preset isEqualToString : AVCaptureSessionPreset640x480]) {
        return CGSizeMake(640, 480);
    }
    
    if ([preset isEqualToString : AVCaptureSessionPreset1280x720]) {
        return CGSizeMake(1280, 720);
    }
    
    if ([preset isEqualToString : AVCaptureSessionPreset1920x1080]) {
        return CGSizeMake(1920, 1080);
    }
    
    return CGSizeMake(0, 0);
}

-(BOOL)setSize:(CGSize)size
{
    if (!opened_) {
        return FALSE;
    }
    CGSize optSize = [self getOptimalPreviewSize : size];
    for (NSString* preset in self.allSupportedPreviewSize) {
        if (([preset isEqualToString : AVCaptureSessionPreset640x480] && optSize.width == 640 && optSize.height == 480) ||
            ([preset isEqualToString : AVCaptureSessionPreset1280x720] && optSize.width == 1280 && optSize.height == 720) ||
            ([preset isEqualToString : AVCaptureSessionPreset1920x1080] && optSize.width == 1920 && optSize.height == 1080)) {
            if (optSize.width != self.previewSize.width || optSize.height != self.previewSize.height) {
                [self.captureSession setSessionPreset : preset];
                self.previewSize = optSize;
            }
            easyar_Vec2I *cameraSize = [easyar_Vec2I create:@[@(self.previewSize.width), @(self.previewSize.height)]];
            CameraParams = [easyar_CameraParameters createWithDefaultIntrinsics: cameraSize cameraDeviceType:easyar_CameraDeviceType_Back cameraOrientation:90];
            return TRUE;
        }
    }
    return FALSE;
}

-(CGSize)getOptimalPreviewSize:(CGSize)size
{
    long area = size.width * size.height;
    
    CGSize res = CGSizeMake(0, 0);;
    if (self.allSupportedPreviewSize.count > 0)
        res = [self getSupportedSize : 0];
    long minAreaDiff = std::numeric_limits<long>::max();
    for (int i = 0; i < [self getNumSupportedSize]; ++i) {
        CGSize sizei = [self getSupportedSize : i];
        long areaDiff = std::abs(sizei.width * sizei.height - area);
        
        if (areaDiff < minAreaDiff) {
            minAreaDiff = areaDiff;
            res = sizei;
        }
    }
    return res;
}

#pragma mark -
#pragma mark Helper Methods
- (AVCaptureDevice*)cameraWithPosition:(AVCaptureDevicePosition)position
{
    NSArray *devices = [AVCaptureDevice devicesWithMediaType : AVMediaTypeVideo];
    for (AVCaptureDevice *device in devices) {
        if ([device position] == position) {
            return device;
        }
    }
    return nil;
}

-(void)addVideoDataOutput
{
    self.captureOutPut = [[AVCaptureVideoDataOutput alloc] init];
    self.captureOutPut.alwaysDiscardsLateVideoFrames = YES;
    
    dispatch_queue_t queue;
    queue = dispatch_queue_create("cn.easyar.engine", DISPATCH_QUEUE_SERIAL);
    [self.captureOutPut setSampleBufferDelegate : self queue : queue];
    
    NSString *key = (NSString *)kCVPixelBufferPixelFormatTypeKey;
    NSNumber *value = [NSNumber numberWithUnsignedInt : kCVPixelFormatType_420YpCbCr8BiPlanarFullRange];
    NSDictionary *settings = @{key:value};
    [self.captureOutPut setVideoSettings : settings];
    
    [self.captureSession addOutput : self.captureOutPut];
}

-(void)captureOutput:(AVCaptureOutput *)captureOutput
didOutputSampleBuffer : (CMSampleBufferRef)sampleBuffer
     fromConnection : (AVCaptureConnection *)connection
{
    (void)captureOutput;
    (void)connection;
    CVImageBufferRef imageBuffer = CMSampleBufferGetImageBuffer(sampleBuffer);
    
    CVPixelBufferLockBaseAddress(imageBuffer, kCVPixelBufferLock_ReadOnly);
    
    size_t width = CVPixelBufferGetWidth(imageBuffer);
    size_t height = CVPixelBufferGetHeight(imageBuffer);
    
    uint8_t *yAddr = (uint8_t*)CVPixelBufferGetBaseAddressOfPlane(imageBuffer, 0);
    uint8_t *uvAddr = (uint8_t*)CVPixelBufferGetBaseAddressOfPlane(imageBuffer, 1);
    
    if(!yAddr || !uvAddr) return;
    
    int ySize = (int)(width * height);
    int uvSize = ySize/2;
    
    easyar_Buffer *buffer = [easyar_Buffer create: (int)(width * height * 3 / 2)];
    [buffer tryCopyFrom:yAddr srcIndex:0 index:0 length:ySize];
    [buffer tryCopyFrom:uvAddr srcIndex:0 index:ySize length:uvSize];
    CVPixelBufferUnlockBaseAddress(imageBuffer, kCVPixelBufferLock_ReadOnly);
    easyar_Image *image = [easyar_Image create: buffer format:easyar_PixelFormat_YUV_NV12 width:(int)width height:(int)height];
    
    easyar_InputFrame * frame = [easyar_InputFrame createWithImageAndCameraParametersAndTemporal:image cameraParameters:CameraParams timestamp:CMTimeGetSeconds(CMSampleBufferGetPresentationTimeStamp(sampleBuffer))];
    [self.sink handle:frame];
};

-(void)dealloc
{
    if (self.captureOutPut != nil)
        [self.captureSession removeOutput: self.captureOutPut];
    [self.captureSession stopRunning];
}

@end

